﻿/*
*Copyright(c) Live2D Inc. All rights reserved.
*このソースコードの使用には、"http://www.live2d.com/eula/live2d-open-software-license-agreement_jp.html"ファイルにあるLive2D Open Software ライセンスが適用されます。
*Use of this source code is governed by a Live2D Open Software license that can be found in the "http://www.live2d.com/eula/live2d-open-software-license-agreement_en.html" file.
*/

using System;
using UnityEngine;
using System.Collections;
using System.Collections.Generic;
using System.Security.AccessControl;
using live2d.Euclid;
using UnityEngine.VR;

[RequireComponent(typeof(Camera))]
public class EuclidVRCamera : MonoBehaviour
{
    [SerializeField]
    private List<EuclidModel> TargetEuclidModel = new List<EuclidModel>();
    [SerializeField]
    public float BothEyeOffset = 0.0f; // Adjust parallax

    #region private field
    private EuclidCameraBackend _euclidCameraBackendLeft = new EuclidCameraBackend(false);
    private EuclidCameraBackend _euclidCameraBackendRight = new EuclidCameraBackend(true);

    private float leftCameraOffset;
    private float rightCameraOffset;
    private float eyeOffset;

    private int _renderingEyeId;
    #endregion

    /// <summary>
    /// シーン開始後にモデルをカメラに追加する
    /// シーン開始前や開始中に呼ばれた場合の動作は未定義
    /// </summary>
    public void AddEuclidModel(EuclidModel model)
    {
        if (TargetEuclidModel.Contains(model) || model == null) { return; }

        TargetEuclidModel.Add(model);
        var left = registerEuclidModel(_euclidCameraBackendLeft, model);
        var right = registerEuclidModel(_euclidCameraBackendRight, model);
        _euclidCameraBackendLeft.InitializeModel(transform, left);
        _euclidCameraBackendRight.InitializeModel(transform, right);
    }

    private EuclidModelBackend registerEuclidModel(EuclidCameraBackend backend, EuclidModel model)
    {
        if (model == null) { return null; }

        var modelBackend = model.AddBackend();
        backend.registerEuclidModel(modelBackend);

        return modelBackend;
    }

    private void Awake()
    {
        // Set ID for shader
        _renderingEyeId = Shader.PropertyToID("RenderingEye");

        foreach (var m in TargetEuclidModel)
        {
            registerEuclidModel(_euclidCameraBackendLeft, m);
            registerEuclidModel(_euclidCameraBackendRight, m);
        }
    }

    public void ResetParallaxOffset()
    {
        var center = InputTracking.GetLocalPosition(VRNode.CenterEye);

        // Get offset both eye's
        leftCameraOffset  = (center - InputTracking.GetLocalPosition(VRNode.LeftEye)).magnitude;
        rightCameraOffset = (center - InputTracking.GetLocalPosition(VRNode.RightEye)).magnitude;
        eyeOffset = (leftCameraOffset + rightCameraOffset) * BothEyeOffset * 0.5f;

    }

    private void OnEnable()
    {
        _euclidCameraBackendLeft.InitializeModel(transform);
        _euclidCameraBackendRight.InitializeModel(transform);
    }

    private void Start()
    {
        ResetParallaxOffset();
    }

    private void Update()
    {
        var leftEyePos   = transform.position + -transform.right * leftCameraOffset;
        var rightEyePos  = transform.position +  transform.right * rightCameraOffset;

        // for Euclid Model update
        _euclidCameraBackendLeft.UpdateEuclidModelParameter(leftEyePos);
        _euclidCameraBackendRight.UpdateEuclidModelParameter(rightEyePos);
    }


    private void LateUpdate()
    {
        var leftEyePos  = transform.position + -transform.right * (leftCameraOffset  + eyeOffset);
        var rightEyePos = transform.position +  transform.right * (rightCameraOffset + eyeOffset);

        // Update Euclid Model Parameter
        _euclidCameraBackendLeft.LateUpdateEuclidModelParameter(leftEyePos);
        _euclidCameraBackendRight.LateUpdateEuclidModelParameter(rightEyePos);
    }


    private int _currentEye = 0;
    private void OnPreRender()
    {
        // Set current eye to shader
        Shader.SetGlobalInt(_renderingEyeId, _currentEye);

        // Update current eye
        _currentEye = 1 - _currentEye;
    }

    private void OnDisable()
    {
        _euclidCameraBackendLeft.Close();
        _euclidCameraBackendRight.Close();
    }

    private void OnDestroy()
    {
        _euclidCameraBackendLeft.Destroy();
        _euclidCameraBackendRight.Destroy();
    }
}
